#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.5

$LastChangedRevision: 6593 $
$LastChangedDate:: 2023-02-10 09:31:42 +0100#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import os
import sys
import re
import glob
import optparse
import json
import stat
import time
import shutil
import warnings
from symspace.osutils.fileutils import ShutilWrapper
from symspace.osutils.jenkins_utils import isBuildServer
from componentspace.utils.image_processing import ImageProcessing
from componentspace.help.component_help import ComponentHelp
import componentspace.osutils.fileutils as cs_fileutils
from componentspace.osutils.paths import ComponentPaths
from componentspace.osutils.svnutils import ComponentSVNUtils

try:
    from lcm.tdata_interface.pyjdata_interface import PyJDataInterface
    from symspace.component import readInterface
    from symspace.gui.qt.form_dialog import FormDialog
    from symspace.pyutils.pkgutil import PkgUtil
    from symspace.pyutils.distpkgutil import DistPkgUtil
    from symspace.osutils.pathutils import Paths
    from symspace.osutils.pathutils import CMDParser
    import symspace.osutils.processutils as pu
    from symspace.wrapper.symwrapper import (SymProject, SymFile, SymDirectory,
                                             SymFunction, SymContainer, SymComponent, SymField)
except ImportError:
    raise Exception('Update SyMSpace to the latest version')

try:
    # noinspection PyUnresolvedReferences
    from magopt.utils import MopFile, ComponentInfo
except Exception as e:
    print(e)


# ToDo: Major code clean-up needed
class Component:
    """
    Components for SyMSpace
    """
    py_pkg_file_format = 'zip'

    # ToDo: create a __init__ method which is compatible with the subclasses (Geometry, etc.)
    #  e.g. create two different classes for the cases comp_file: mandatory/optional
    #  to avoid constant checks if self.name/path/sym_mdl is None
    def __init__(self, comp_file=None):
        """
        Init function

        Parameters
        ----------
        comp_file : str
            SyMSpace file for definition of the component
        """
        if comp_file is None:
            self.name = None  # Name of the component
            self.path = None  # File path to the folder
            self.sym_mdl = None
            self.comp_par = None  # SymProject parameter
            # Todo: remove self.comp_par in future, but currently it is still used in some build_component.py files
        else:
            self.name = os.path.basename(comp_file).replace('.mop', '')
            self.path = os.path.dirname(comp_file)
            self.sym_mdl = SymProject.load(os.path.join(self.path, self.name))
            self.comp_par = self.sym_mdl.getObject('Project')  # type: SymProject

    @property
    def base_directory(self) -> str:
        """Return the directory where the base/root component is located,
            this may be different from the component path for some component subclasses (Geometry, Material, ...)
        """
        return self.path

    @classmethod
    def fromCMDParser(cls) -> 'Component':
        """
        Commandline parser to create instance of this class

        Commandline Arguments
        ---------------------

        --fullBuild : int
            If --fullBuild is 1 all components will be built,
            if --fullBuild is 0 (Default) only new components will be built.

        --startDir : str
            The search for component starts from this directory (Optional).
            Default is the ComponentSpace root directory.

        Returns
        -------
        builder : ComponentBuilder
        """

        parser = optparse.OptionParser("usage: %prog [options]")
        parser.add_option("-g", "--usegui", dest="use_gui",
                          default='1', type="int",
                          help="Use graphical user interface")
        parser.add_option("-n", "--name", dest="comp_name",
                          default='MyComponent', type="string",
                          help="Componentname")
        parser.add_option("-p", "--path", dest="comp_path", default=Paths.getComponentSpaceRoot(),
                          type="string", help="The search for components starts from this directory, "
                                              "default is the ComponentSpace root directory.")
        args = CMDParser.jepSysArgs()
        (options, args) = parser.parse_args(args)
        # if len(args) != 0:
        #     parser.error("incorrect number of arguments")

        comp = cls()

        use_gui = bool(options.use_gui)

        if not use_gui:
            comp.name = options.comp_name
            comp.path = os.path.join(CMDParser.jepArgPath(options.comp_path), comp.name)

            print('Name: {}'.format(comp.name))
            print('Path: {}'.format(comp.path))

        return comp

    @staticmethod
    def printH1(value):
        print('\n{}\n{}\n'.format(value, len(value) * '='))

    @staticmethod
    def printH2(value):
        print('\n{}\n{}\n'.format(value, len(value) * '-'))

    @staticmethod
    def printH3(value):
        print('\n{}\n{}\n'.format(value, len(value) * '~'))

    def writeVersionFile(self) -> str:
        version_str = str(self.sym_mdl.getObject('Info.Version').getData())

        regex_pattern = r"^([\d]+\.[\d]+)$"
        match_obj = re.search(regex_pattern, version_str)

        if match_obj is None:
            raise ValueError('Version must be in the format: "Major.Minor", got "{}" instead!'.format(version_str))

        version_txt = ComponentPaths.getVersionFile(self.path)

        with open(version_txt, 'w') as fh:
            fh.write(version_str)
        return version_str

    def generateHelpFiles(self, comp_cont: SymContainer):
        """
        Generate help files

        Parameters
        ----------
        comp_cont : SymContainer
        """
        rst_path = os.path.join(self.path, 'rst')

        param_file = os.path.join(rst_path, 'Param.txt')
        param_detail_file = os.path.join(rst_path, 'Paramdetail.txt')

        '''get the original type of the container'''
        param_id = comp_cont.getIdentifier()
        param = comp_cont.getObject(param_id)

        if isinstance(param, SymFunction):
            ''' the component consists only of a function'''

            par_in_lst = readInterface(param.getImports())
            par_out_lst = readInterface(param.getExports())

            param_in_file = os.path.join(rst_path, 'ParamIn.txt')
            param_out_file = os.path.join(rst_path, 'ParamOut.txt')

            '''write input parameters'''
            with open(param_in_file, "w") as fh:

                header = '    '.join(['.. csv-table:: List of input parameters\n',
                                      ':header: "Parameter", "Description", "Info"\n',
                                      ':widths: 5,20,5 \n\n'])
                lines = []
                for item in par_in_lst:
                    if '"' in item['description']:
                        print('-' * 100)
                    lines.append('    "{0}", "{1}", "{2}"\n'.format(item['name'], item['description'].replace('"', ''),
                                                                    item['info'].replace('"', '')))

                fh.write(header)
                fh.writelines(lines)

            '''write output parameters'''
            with open(param_out_file, "w") as fh:
                header = '    '.join(['.. csv-table:: List of output parameters\n',
                                      ':header: "Parameter", "Description", "Info"\n',
                                      ':widths: 5,20,5 \n\n'])
                lines = []
                for item in par_out_lst:
                    lines.append('    "{0}", "{1}", "{2}"\n'.format(item['name'], item['description'].replace('"', '')
                                                                    , item['info'].replace('"', '')))

                fh.write(header)
                fh.writelines(lines)

        '''get component parameters'''
        par_lst = []
        base_identifier = comp_cont.getIdentifier() + '.'
        for param in comp_cont.getParameters('SymParameter'):
            item = {}
            if param.getComment().strip().startswith('#'):
                item['name'] = param.getIdentifier().replace(base_identifier, '')
                item['description'] = param.getDescription()
                if param.getComment().strip().startswith('##') or param.getComment().strip().startswith('#!'):
                    item['info'] = 'fixed'
                else:
                    item['info'] = 'output' if param.getReadOnly() else 'input'

                '''remove # from comment'''
                item['comment'] = param.getComment().replace('#', '').replace('!', '').strip()
                par_lst.append(item)

        '''write component parameters'''
        with open(param_file, "w") as fh:
            header = '    '.join(['.. csv-table:: List of Component parameters\n',
                                  ':header: "Parameter", "Description", "Info"\n',
                                  ':widths: 5,20,5 \n\n'])
            lines = []
            for item in par_lst:
                if item['comment'] == '':
                    lines.append('    "{0}", "{1}", "{2}"\n'.format(item['name'], item['description'], item['info']))
                else:
                    lines.append('    ":ref:`{0} <{3}_{0}>`","{1}", "{2}"\n'.format(item['name'], item['description'],
                                                                                    item['info'], self.name))

            fh.write(header)
            fh.writelines(lines)

        '''write details of component parameters'''
        with open(param_detail_file, "w") as fh:
            lines = []
            for item in par_lst:
                if item['comment'] != '':
                    lines.append('.. _{0}_{1}:\n\n'.format(self.name, item['name']))
                    lines.append('*{0}:*\n    {1}\n\n'.format(item['name'], item['comment']))

            fh.writelines(lines)

    @staticmethod
    def openCreateComponentGUI(templates: tuple = None) -> dict:
        """
        Graphical User Interface for component creation

        Returns
        -------
        component_data : dict
        """
        if templates is None:
            templates = ()

        cs_root = Paths.getComponentSpaceRoot()

        '''find UserGroups'''
        libs = os.listdir(os.path.join(cs_root, '_UserGroups'))
        if 'Open' in libs:
            '''set Open as first (=> default) entry'''
            libs.remove('Open')
            libs.insert(0, 'Open')

        if len(templates) == 0:
            '''find templates'''
            templates_files = glob.glob(os.path.join(cs_root, '**', 'template_build_component.py'), recursive=True)
            templates = [os.path.dirname(file) for file in templates_files]
            templates.sort(reverse=True)
            default_template = os.path.join(cs_root, '_template')
            if default_template != templates[0]:
                '''set default as first entry'''
                templates.remove(default_template)
                templates.insert(0, default_template)
        else:
            for template in templates:
                if not os.path.isdir(template):
                    raise NotADirectoryError('The given template is not a directory: "{}"'.format(template))

        form_title = 'Create new Component'
        form_elements = [('Component name', 'line', 'MyComponent'),
                         ('Select directory', 'browse', Paths.getComponentSpaceRoot()),
                         ('Select template', 'combobox', templates),
                         ('Author', 'line', 'name@lcm.at'),
                         ('Tags', 'line'),
                         ('Library', 'combobox', libs)]
        form_result = FormDialog.run(form_title, form_elements)

        if form_result['accept']:
            return form_result
        else:
            print('Cancel -> Closing GUI and aborting process')
            sys.exit()  # exit program

    def createComponent(self, template: str = None, component_data: dict = None) -> str:
        """
        Create component definition file form template

        Parameters
        ----------
        template : str
        component_data : dict

        Returns
        -------
        fname : str
            File name of component
        """
        self.printH1('Create Component')

        if self.path is None and component_data is None:
            '''Open dialog if name is not defined'''
            templates = (template, ) if template is not None else None
            component_data = self.openCreateComponentGUI(templates)
            self.name = component_data['Component name']
            self.path = os.path.join(component_data['Select directory'], self.name)
            template = component_data['Select template']
        elif self.path is None and component_data is not None:
            self.name = component_data['Component name']
            self.path = os.path.join(component_data['Select directory'], self.name)
        else:
            raise Exception('Unexpected chase!')

        if template is None:
            parent_dir = os.path.dirname(self.path)
            alt_template_dir = os.path.join(parent_dir, '_template')

            if os.path.exists(alt_template_dir):
                '''Use template if exists in target directory'''
                template = alt_template_dir
                print('Using alternative component template: {}'.format(template))
            else:
                '''Use standard template'''
                template = os.path.join(Paths.getComponentSpaceRoot(), '_template')
        else:
            print('Using component template: {}'.format(template))

        if os.path.exists(template):
            '''Copy template files'''
            if not os.path.exists(self.path):
                os.makedirs(self.path)
            else:
                print(self.path)
                raise Exception('Path already exists')

            cs_fileutils.copyTree(template, self.path)
            os.rename(os.path.join(self.path, 'Template.mop'),
                      os.path.join(self.path, self.name + '.mop'))
            if os.path.exists(os.path.join(self.path, 'Template.mop.Files')):
                os.rename(os.path.join(self.path, 'Template.mop.Files'),
                          os.path.join(self.path, self.name + '.mop.Files'))

            '''build component'''
            if os.path.isfile(os.path.join(self.path, 'build_component.py')):
                os.remove(os.path.join(self.path, 'build_component.py'))
            os.rename(os.path.join(self.path, 'template_build_component.py'),
                      os.path.join(self.path, 'build_component.py'))
            cs_fileutils.replaceInFile(os.path.join(self.path, 'build_component.py'), '__compdef__', self._strMopFile())

            '''Rename index.txt'''
            display = self.name.replace('_', ' ')
            if os.path.isfile(os.path.join(self.path, 'index.txt')):
                os.rename(os.path.join(self.path, 'index.txt'), os.path.join(self.path, 'index.rst'))
            cs_fileutils.replaceInFile(os.path.join(self.path, 'index.rst'), '__ComponentTitle__',
                                       display + '\n' + '=' * len(display))

            '''Open SyMSpace Project'''
            prj = SymProject.load(os.path.join(self.path, self.name + '.mop'))
            prj['Info.Name'].setData(self.name)
            if len(prj['Info.FunctionName'].getData()) == 0:
                prj['Info.FunctionName'].setData(self.name)

            if prj['Info.DisplayName'] is not None:
                prj['Info.DisplayName'].setData(display)

            '''fill fields with GUI inputs'''
            if component_data is not None:
                prj['Info.Author'].setData(component_data['Author'])
                prj['Info.Tags'].setData(component_data['Tags'])
                prj['Filter.UserGroup'].setData(component_data['Library'])

            if self.path.startswith(Paths.getComponentSpaceRoot()):
                web_path = self.path.replace(Paths.getComponentSpaceRoot(), '') \
                    .replace('\\' + self.name, '').replace('\\', '.')
                if web_path.startswith('.'):
                    web_path = web_path[1:len(web_path)]
                prj['Info.WebPath'].setData(web_path)
            prj.save()
            prj.close()

            print('Component project saved to: {}'.format(self.path))
            file_name = os.path.join(self.path, self.name + '.mop')
            return file_name
        else:
            raise Exception('Template path does not exist!')

    def buildHelp(self, version_str: str):
        """
        Build the component help
        """
        self.printH1('Build Help')
        comp_help = ComponentHelp(self.name, self.path)
        if comp_help.checkRoot():
            '''preprocess'''
            comp_help.copyConfFiles()
            comp_help.build()

            '''postprocess'''
            comp_help.cleanUpFolders()
            # comp_help.writeCleanHtml()  # Todo: maybe this needs to be removed in future, to use sphinx footers etc.
            comp_help.copyComponentImage()
            if version_str < '1.0' and \
                    os.path.isfile(os.path.join(self.path, 'help', 'Component.png')):
                img = ImageProcessing(os.path.join(self.path, 'help', 'Component.png'))
                logo = ImageProcessing(os.path.join(Paths.getComponentSpaceRoot(), '_template', 'rst', 'Component.png'))
                img.merge(logo)
                img.save()

    def buildComponent(self, keep_mopc: bool = False) -> str:
        """
        Build component files

        Parameters
        ----------
        keep_mopc : bool

        Returns
        -------
        dst_file : str
        """
        self.printH1('Build Component')

        if self.sym_mdl is None:
            file_model = os.path.join(self.path, self.name + '.mop')
            if os.path.isfile(file_model):
                self.sym_mdl = SymProject.load(file_model)
            else:
                raise Exception('Model {} file does not exist!'.format(file_model))

        name_comp = self.sym_mdl.getObject('Name').getData()
        path_web = self.sym_mdl.getObject('WebPath').getData()
        tags_comp = self.sym_mdl.getObject('Tags').getData()
        # path_hlp = self.sym_mdl.getObject('Help').getData()
        embed_scripts = self.sym_mdl.getObject('EmbedScripts').getData()

        if self.sym_mdl.getObject('FunctionName') is None:
            name_fnc = name_comp
        else:
            name_fnc = self.sym_mdl.getObject('FunctionName').getData()

        '''Create new component file'''
        fname_tmp_mop = os.path.join(self.path, self.name + '_.mop')
        fname_tmp_mopc = os.path.join(self.path, self.name + '_.mopc')

        p_tmp = SymProject.new('Component_Temp__', '', fname_tmp_mop)
        comp_list = []
        component = self.__getComponent(p_tmp, name_fnc, comp_list)
        p_folder = p_tmp.getAllFilesDirectory()

        '''Infos'''
        self.printH2('Write Version Files')
        version_str = self.writeVersionFile()
        revision, _ = ComponentSVNUtils.writeRevisionFile(self.base_directory)
        full_version_str = '{}.{:d}'.format(version_str, revision)
        print('Version: {}, Revision: {:d}'.format(version_str, revision))

        if embed_scripts:
            '''public_library_path
               could be used for code, which is available for all components of a project e.g. packages
               therefore it should point to a folder which is not component specific e.g. *.mop.Files/_Scripts
               
               
               private_library_path
               could be used for code, which is available only for this specific component e.g. modules
               therefore it should point to a folder which is component specific e.g. *.mop.Files/_Comp/xy
             '''
            self.printH2('Embed Scripts')
            public_library_path = os.path.join(p_folder, '_Script')
            private_library_path = os.path.join(p_folder, '_Script')  # Todo: change to component folder

            for path in [public_library_path, private_library_path]:
                if not os.path.exists(path):
                    os.makedirs(path)

            '''add Matlab scripts'''
            self.__addMatlabFiles(component, public_library_path)
            '''add Python scripts'''
            self.__addPythonFiles(component, public_library_path, private_library_path)

            '''add additional scripts'''
            self.__addScripts(component, public_library_path)

        '''add Files'''
        self.__addFiles(component, p_folder)
        '''add Component Scripts'''
        self.__addComponentScripts(component)
        '''add Help'''
        self.__addHelp(component, version_str)
        '''add Docu'''
        self.__addDocu(component)
        '''Remove tags'''
        self.__removeTags(component)

        '''Infos'''
        self.printH2('Edit Infos')
        full_webpath = '{}.{}'.format(path_web, name_comp)

        infos = ComponentInfo(MopFile(fname_tmp_mopc))  # ToDo: Change to wrapper
        infos.setName(full_webpath)
        infos.setTags(tags_comp)

        '''Write Component information'''
        self.printH2('Write Component Information')
        self.__writeComponentInfo(component, full_webpath, full_version_str)
        # component.getContainer().setComment(full_webpath)

        '''Write component'''
        p_tmp.javaObj.exportComponentList(MopFile(fname_tmp_mopc), name_fnc, comp_list, infos)

        '''Find error in project p_tmp'''
        self.printH2('Find Errors')
        err_str = p_tmp.getEventLog()
        pos = err_str.find('Error')
        if pos >= 0:
            print(err_str[pos: err_str.find('\\', pos)])

        '''Reset errors in Component and update component'''
        prj_comp = SymProject.load(fname_tmp_mopc)
        for par in prj_comp.getParameters('SymContainer'):
            par.setError(False)

            if par.isComponent():
                if par.getComponent().getName() == full_webpath:
                    continue
                self.printH3(f'Update {par.getName()}')
                try:
                    par.updateComponent()
                except:
                    print(f'    -- {par.getName()} update not possible')

        '''Get help path'''
        if prj_comp.isComponent():
            pth = prj_comp.javaObj.getComponent().getHelpFolder().toString()
            help_dir = pth.split(os.sep)[-2]
        elif prj_comp.getParam()[0].isComponent():
            pth = prj_comp.getParam()[0].javaObj.getComponent().getHelpFolder().toString()
            help_dir = pth.split(os.sep)[-2]
        else:
            # Todo: what is this case?
            help_dir = ''

        '''Get structure and unresolved interface'''
        prj_info = dict()
        prj_info['Project'] = Component.getStructure(prj_comp)
        if self.sym_mdl.getObject('Info.FunctionName'):
            prj_info['Project']['name'] = self.sym_mdl.getObject('Info.FunctionName').getData()
        else:
            prj_info['Project']['name'] = self.sym_mdl.getObject('Info.Type').getData()
        prj_info['Unresolved'] = Component.getUnresolvedInterface(prj_comp)
        prj_info['Dependencies'] = self.getDependencies(prj_comp, full_webpath)

        prj_comp.save()
        prj_comp.close()

        '''Write JSON file'''
        self.printH2('Write JSON file')
        '''change component webpath for json file'''
        self.sym_mdl['Info.WebPath'].setData(full_webpath)
        self.writeJSON(fname_tmp_mopc, help_dir, **prj_info)
        '''revert webpath to original'''
        self.sym_mdl['Info.WebPath'].setData(path_web)

        if not keep_mopc:
            dst_file = self.moveCompFiles()
        else:
            dst_file = None  # Todo: implement return of dst_file (dst_file is used in UnitTests)

        # '''Remove rst/*.txt'''
        # for fh in os.listdir(os.path.join(self.path, 'rst')):
        #     if fh.endswith('.txt'):
        #         os.remove(os.path.join(self.path, 'rst', fh))
        # #
        # p_tmp.save()

        self.printH2('Finalize build')
        p_tmp.close()

        return dst_file

    def __getComponent(self, p_tmp: SymProject, name_fnc: str, comp_list: list) -> SymComponent:
        """

        Parameters
        ----------
        p_tmp
        name_fnc

        Returns
        -------
        SymComponent
        """

        def getContainer(_params):
            if len(_params) == 1:
                container = _params[0].cloneParam()
                p_tmp.addParam(container)
                comp_list.append(name_fnc)
            else:
                container = p_tmp
                for _param in _params:
                    p_tmp.addParam(_param.cloneParam())
                    comp_list.append(name_fnc + '.' + _param.getName())
            return container

        prj_params = self.sym_mdl.getObject('Project').getParam()

        if len(prj_params) == 0:
            raise Exception('Container Project is empty, please add Component here')

        if isinstance(prj_params[0], SymFile):
            '''Load a template/project if the first element is a SymFile'''
            self.printH2('Loading template')
            template_file = prj_params[0].getData()
            print('Template file: {}'.format(template_file))
            prj_ext = SymProject.load(template_file)
            comp_container = getContainer(prj_ext.getParam())
            prj_ext.close()

        else:
            comp_container = getContainer(prj_params)

        comp_container.setName(name_fnc)
        comp_container.addComponentProperty()
        component = comp_container.getComponent()
        return component

    @classmethod
    def __removeTags(cls, component: SymComponent):
        """remove Tags"""
        cls.printH2('Removing comments')
        comp_container = component.getContainer()
        for param in comp_container.getParameters('SymParameter'):
            if param.getComment().startswith('##') or param.getComment().replace('#', '').startswith('!'):
                param.setLock(True)
            if param.getComment().startswith('#') or param.getComment().startswith('!'):
                param.setComment('')

    def __addFiles(self, component: SymComponent, p_folder: str):
        """add Files"""
        params = self.sym_mdl.getObject('Files').getParam()
        if len(params) > 0:
            self.printH2('Add Files')
            comp_container = component.getContainer()
            comp_files = comp_container.getParameters('SymFile')
            comp_dirs = comp_container.getParameters('SymDirectory')
            c_folder = os.path.join(p_folder, '_CompFiles')

            print('Copying files and directories ...')
            for param in params:
                if isinstance(param, SymFile):
                    file_path = param.getData()
                    self.__addSyMFile(component, c_folder, comp_files, file_path)

                elif isinstance(param, SymDirectory):
                    sdir = param.getData()
                    self.__addSyMDirectory(component, p_folder, c_folder, comp_files, comp_dirs, sdir)

    def __addHelp(self, component: SymComponent, version_str: str):
        """add Help"""
        self.printH2('Add Help')
        comp_container = component.getContainer()
        hlp_comp = component.getHelpFolder()
        self.generateHelpFiles(comp_container)
        self.buildHelp(version_str)
        cs_fileutils.copyTree(os.path.join(self.path, 'help'), hlp_comp)

    def __addDocu(self, component: SymComponent):
        params = self.sym_mdl.getObject('Docu').getParam()
        if len(params) > 0:
            self.printH2('Add Docu')
            print('Copying documentation files and directories ...')

            hlp_comp = component.getHelpFolder()
            d_folder = hlp_comp.replace('Help', 'Docu')
            os.makedirs(d_folder)
            for param in params:
                if isinstance(param, SymFile):
                    ShutilWrapper.copy2(param.getData(), d_folder)
                elif isinstance(param, SymDirectory):
                    sdir = param.getData()
                    cs_fileutils.copyTree(sdir, os.path.join(d_folder, os.path.basename(os.path.normpath(sdir))))

    def __addScripts(self, component: SymComponent, script_folder: str):
        script_param = self.sym_mdl.getObject('Packages.Scripts')
        if script_param:
            params = script_param.getParam()
            if len(params) > 0:
                self.printH3('Add Scripts')
                if not os.path.isdir(script_folder):
                    os.makedirs(script_folder)

                print('Copying Script files ...')
                for param in params:
                    if isinstance(param, SymFile):
                        src_file = param.getData()
                        dst_file = os.path.join(script_folder, os.path.basename(src_file))
                        self.__linkFile(component, src_file, dst_file, True)

    def __addComponentScripts(self, component: SymComponent):
        script_param = self.sym_mdl.getObject('Script')
        # Todo: check if this field "Script" even exists in any components
        #  or if it is meant to be "Packages.Scripts".
        #  Compare with the method "__addScripts"

        if script_param:
            self.printH2('Add Component Scripts')
            hlp_comp = component.getHelpFolder()
            script_folder = hlp_comp.replace('Help', 'Script')
            if not os.path.isdir(script_folder):
                os.makedirs(script_folder)

            print('Copying Script files ...')
            for param in script_param.getParam():
                if isinstance(param, SymFile):
                    src_file = param.getData()
                    dst_file = os.path.join(script_folder, os.path.basename(src_file))
                    self.__linkFile(component, src_file, dst_file, True)

    def __writeComponentInfo(self, component: SymComponent, name: str, full_version_str: str):
        """ Write Component information as JSON string in Comment field"""

        def getSubparameters(startparam: SymContainer):
            for p in startparam.getParam():
                param.append(p.getIdentifier())
                if isinstance(p, SymContainer) and not p.isComponent():
                    getSubparameters(p)

        info = dict(name=name, version=full_version_str)
        param = []
        getSubparameters(component.getContainer())
        for idx, p in enumerate(param):
            param[idx] = p.replace(component.getContainer().getIdentifier() + '.', '')

        info['parameter'] = param
        # jstr = json.dumps(info)
        component.getContainer().setComment(name)
        for key in info:
            component.setJSONValue(key, info[key])

    @staticmethod
    def __linkFile(component: SymComponent, src_file: str, dst_file: str, relink: bool):
        comp_linked_files = component.listLinkedFiles()
        comp_files = component.getContainer().getParameters('SymFile')  # type: list[SymFile]

        if dst_file not in comp_linked_files:
            if not os.path.isfile(dst_file):
                '''destination file does not exist'''
                ShutilWrapper.copy2(src_file, dst_file)
                component.addLinkedFile(dst_file)
                print('... successfully copied {}'.format(dst_file))
            elif os.path.isfile(dst_file) and not os.path.samefile(src_file, dst_file):
                '''destination file does exist but is different from source file'''
                ShutilWrapper.copy2(src_file, dst_file)
                component.addLinkedFile(dst_file)
                print('... successfully copied {}'.format(dst_file))
            else:
                print('... file already exists {}'.format(dst_file))

        if os.path.isfile(dst_file) and relink:
            for sym_file in comp_files:
                '''re-link fields of type SymFile, containing the file path'''
                if sym_file.getData() == src_file:
                    sym_file.setData(dst_file)
                    print('... ... re-linking file {}'.format(src_file))

    @classmethod
    def __addMatlabFiles(cls, component: SymComponent, library_path: str):
        mat_functions = component.getContainer().getParameters('SymMatlabFunction')  # type: list[SymFunction]

        if len(tuple(mat_functions)) > 0:
            cls.printH3('Add Matlab Files')

            mscript_files = []
            for fnc in mat_functions:
                im_assign = fnc.getImportAssignments()
                if 'mscript' in im_assign:
                    mscript_field = fnc.getObject(im_assign['mscript'])
                    if isinstance(mscript_field, SymFile):
                        mscript_file = mscript_field.getData()
                        if mscript_file.endswith('.m') and os.path.isfile(mscript_file):
                            mscript_files.append(mscript_file)
                        else:
                            raise NotImplementedError(
                                'mscript-file is not an existing m-file! => {}'.format(mscript_file))
                    elif isinstance(mscript_field, SymField):
                        raise NotImplementedError(
                            'mscript-field is not a SymFile!  => {}'.format(mscript_field.getName()))
                    else:
                        raise NotImplementedError(
                            'Problems with mscript-field of MatlabFunction!  => {}'.format(fnc.getName()))
                else:
                    '''
                    Supports only SyMSpace\\Library\\_General\\Tools\\Functions\\Matlab\\@FncMatlabScript\\FncMatlabScript.m
                    '''
                    print('This Matlab Function is currently not supported!')

            copy_mat_files = []
            copy_mat_files.extend(mscript_files)

            for src_file in set(copy_mat_files):
                '''copy and link m-files to component'''
                dst_file = os.path.join(library_path, os.path.basename(src_file))
                cls.__linkFile(component, src_file, dst_file, True)

    def __addPythonFiles(self, component: SymComponent, public_library_path: str, private_library_path: str):
        py_functions = component.getContainer().getParameters('SymPythonFunction')

        if len(tuple(py_functions)) > 0:
            self.printH3('Add Python files')
            fnc_files = []

            for fnc in py_functions:
                cp = fnc.getClassPath()
                module_name = fnc.getModuleName()

                if '.' not in module_name and cp != '':
                    '''standard py-file exists'''
                    if hasattr(fnc, 'getModulePath'):
                        py_file = fnc.getModulePath()
                    else:
                        py_file = os.path.join(fnc.getClassPath(), fnc.getModuleName() + '.py')
                    fnc_files.append(py_file)
                    fnc.setClassPath(private_library_path)
                elif '.' in module_name:
                    '''py-file is located in a package'''
                    fnc.setClassPath(private_library_path)
                else:
                    raise NotImplementedError('Unknown case occurred, while handling "SymPythonFunctions"!')

            copy_modules, copy_pkgs = self.collectPythonFiles(fnc_files, auto_collect=False)

            copy_py_files = []
            copy_py_files.extend(copy_modules)
            copy_py_files.extend(copy_pkgs)

            for src_file in set(copy_py_files):
                '''copy and link python packages/modules to component'''
                if src_file in copy_modules:
                    dst_file = os.path.join(private_library_path, os.path.basename(src_file))
                else:
                    dst_file = os.path.join(public_library_path, os.path.basename(src_file))
                self.__linkFile(component, src_file, dst_file, False)

    def moveCompFiles(self, keep_mopc: bool = False) -> str:
        """Move component file to correct folder"""

        self.printH2('Move Component files')
        path_web = self.sym_mdl.getObject('WebPath').getData()
        user_group = self.sym_mdl.getObject('Filter.UserGroup').getData()

        if user_group is None or user_group == '':
            raise ValueError('UserGroup is not defined!')
        if path_web is None or path_web == '':
            raise ValueError('Webpath is not defined!')

        cs_root = Paths.getComponentSpaceRoot()
        dst_dir = os.path.join(cs_root, '_UserGroups', user_group, 'Components')

        '''keep files if user group does not exist'''
        if not os.path.isdir(dst_dir):
            print(f'User group {user_group} does not exist')
            dst_dir = os.path.split(self.path)[0]

        mopc_name = '{}_{}.mopc'.format(path_web.replace('.', '_'), self.name)
        dst_mopc = os.path.join(dst_dir, mopc_name)
        dst_json = dst_mopc.replace('.mopc', '.json')

        tmp_mop = os.path.join(self.path, self.name + '_.mop')
        tmp_mopc = os.path.join(self.path, self.name + '_.mopc')
        tmp_json = tmp_mopc.replace('_.mopc', '_.json')

        print('Copying component file to: {}'.format(dst_mopc))

        for src_file, dst_file in ((tmp_mopc, dst_mopc), (tmp_json, dst_json)):
            if os.path.isfile(dst_file):
                os.chmod(dst_file, stat.S_IWRITE)
                os.remove(dst_file)

            if keep_mopc:
                ShutilWrapper.copy2(src_file, dst_file)
            else:
                shutil.move(src_file, dst_file)

        '''Delete Temp files, clean up files'''
        print('Clean up files ...')
        for mop_file in (tmp_mop, tmp_mopc):
            try:
                ShutilWrapper.rmtree(mop_file + '.Files', onerror=cs_fileutils.on_rm_error)
            except:
                print('Temporary component files cannot be deleted')

        '''Remove rst/_static'''
        try:
            ShutilWrapper.rmtree(os.path.join(self.path, 'rst', '_static'), onerror=cs_fileutils.on_rm_error)
        except:
            print('rst/_static folder cannot be deleted')
        return dst_file

    @staticmethod
    def collectPythonDependencies(prj: SymProject, fnc_files: list = None, auto_collect: bool = False) -> tuple:
        """
        Collect required Python package and module dependencies of the component project

        Parameters
        ----------
        prj : SymProject
            Current component project
        fnc_files : list
            List of ParamPythonFunction files used in the component
        auto_collect: bool
            Automatically collect required dependencies (modules, packages) of the "fnc_files"

        Returns
        -------
        copy_modules : list
            Modules, which the components depends on and need to be copied into the component
        copy_pkgs : dict
            Packages, which the components depends on and need to be copy into the component
        """
        if fnc_files is None:
            fnc_files = []
            component = prj.getObject('Project')
            if isinstance(component, SymContainer):
                py_functions = component.getParameters('SymPythonFunction')
                if len(tuple(py_functions)) > 0:
                    for fnc in py_functions:
                        '''Check if Python module'''
                        if '.' not in fnc.getModuleName():
                            if hasattr(fnc, 'getModulePath'):
                                py_file = fnc.getModulePath()
                            else:
                                py_file = os.path.join(fnc.getClassPath(), fnc.getModuleName() + '.py')
                            fnc_files.append(py_file)

        ''' Add Package path to search path '''
        std_pkgs_path = os.path.join(Paths.getComponentSpaceRoot(), 'src', 'packages')
        std_pkgs = PkgUtil.findSrcPackages([std_pkgs_path, ])

        copy_modules = fnc_files.copy()
        copy_pkgs = dict()
        excplit_pkgs = {}
        excplit_modules = []

        param_pkgs = prj.getObject('Packages')
        param_scripts = prj.getObject('Packages.Scripts')

        if isinstance(param_pkgs, SymContainer):
            '''Collect explicitly defined packages and modules '''
            for param in param_pkgs.getParam():
                if isinstance(param, SymField):
                    """ Collect explicitly defined packages from "%ComponentSpace%/src/packages"

                    +------------+----------------+
                    | Field Name | Field Data     |
                    +============+================+
                    | pkg1       | 'lcm_motorbox' |
                    +------------+----------------+
                    | pkg2       | 'lcm_math'     |
                    +------------+----------------+
                    """
                    pkg_name = param.getData()
                    pkg_name_split = pkg_name.split('.')
                    if len(pkg_name_split) > 1:
                        pkg_name, pkg_name_old = pkg_name_split[0], pkg_name
                        warnings.warn('Including subpackages is deprecated! Changing {} to {}!'.
                                      format(pkg_name_old, pkg_name))

                    if pkg_name not in excplit_pkgs:
                        pkg_dir = std_pkgs[pkg_name]
                        excplit_pkgs[pkg_name] = pkg_dir
                elif isinstance(param, SymDirectory):
                    """ Collect explicitly defined packages from an arbitrary path

                    +----------------------+-----------------------------+
                    | Directory Field Name | Directory Field Data        |
                    +======================+=============================+
                    | pkg1                 | MyComponent/src/my_package1 |
                    +----------------------+-----------------------------+
                    | pkg2                 | MyComponent/src/my_package2 |
                    +----------------------+-----------------------------+
                    """
                    pkg_dir = param.getData()
                    pkg_name = os.path.basename(pkg_dir)
                    if pkg_name not in excplit_pkgs:
                        excplit_pkgs[pkg_name] = pkg_dir
                elif param == param_scripts:
                    pass
                else:
                    raise Exception('Parameter "{}" must be a SymField, SymFile or SymDirectory'
                                    .format(param.getIdentifier()))

            if isinstance(param_scripts, SymContainer):
                '''Collect explicitly defined modules 

                +-----------------+-------------------------------+
                | File Field Name | File Field Data               |
                +=================+===============================+
                | module1         | MyComponent/src/my_module1.py |
                +-----------------+-------------------------------+
                | module2         | MyComponent/src/my_module2.py |
                +-----------------+-------------------------------+
                '''
                for param in param_scripts.getParam():
                    if isinstance(param, SymFile):
                        excplit_modules.append(param.getData())
                    elif isinstance(param, SymDirectory):
                        raise Exception('If "{}" is a custom package, move it from "Packages.Scripts" to "Packages"!'
                                        .format(param.getIdentifier()))
                    else:
                        raise Exception('Parameter "{}" must be a SymFile'.format(param.getIdentifier()))

        '''mark explicitly defined files for copy'''
        copy_modules.extend(excplit_modules)
        copy_pkgs.update(excplit_pkgs)

        if auto_collect:

            '''root paths for component pkgs and modules'''
            comp_src_paths = [os.path.dirname(file) for file in fnc_files if
                              not file.startswith(Paths.getSyMSpaceRoot())]

            '''analyze these modules'''
            search_modules = PkgUtil.findModules(comp_src_paths, recursive=True)
            search_modules.extend(excplit_modules)

            '''search for used pkgs and modules in these paths'''
            importable_paths = [std_pkgs_path, ]
            importable_paths.extend(comp_src_paths)

            '''automatically collect module and package dependencies'''
            dependencies = PkgUtil.getDependencies(search_modules, importable_paths=importable_paths)

            for key, path in dependencies.items():
                '''check if dependency is a package directory or module'''
                if os.path.isdir(path) and path not in copy_pkgs:
                    copy_pkgs[key] = path
                elif os.path.isfile(path) and path not in copy_modules:
                    copy_modules.append(path)

        return copy_modules, copy_pkgs

    def collectPythonFiles(self, fnc_files: list, auto_collect: bool) -> tuple:
        """
        Collect required Python packages and modules

        Parameters
        ----------
        fnc_files : list
            List of ParamPythonFunction files used in the component
        auto_collect: bool
            Automatically collect required dependencies (modules, packages) of the "fnc_files"

        Returns
        -------
        copy_unique_modules : list
            Modules to copy into the component
        copy_unique_dist_pkgs : list
            Package distributions to copy into the component
        """

        ''' Add Package path to search path '''
        std_pkgs_path = os.path.join(Paths.getComponentSpaceRoot(), 'src', 'packages')
        pkg_dist_dir = os.path.join(std_pkgs_path, 'py_dist')
        std_pkgs = PkgUtil.findSrcPackages([std_pkgs_path, ])
        copy_dist_pkgs = []
        pkg_versions = {}

        copy_modules, copy_pkgs = self.collectPythonDependencies(self.sym_mdl, fnc_files, auto_collect)

        '''create package distributions'''
        for pkg_name, pkg_path in copy_pkgs.items():
            try:
                pkg_version = PkgUtil.getSrcPackageVersion(pkg_name, pkg_path)
            except Exception as ex:
                # Todo: this is just a workaround for the ComponentSpace installer versions without svn-connection
                #  and should be replaced by a correct solution
                warnings.warn(str(ex))
                pkg_version = '0.0.0'
            pkg_dist = DistPkgUtil.createFromPkgDir(pkg_path, self.py_pkg_file_format, pkg_dist_dir,
                                                    pkg_version=pkg_version, dry_run=True)

            if os.path.isfile(pkg_dist):
                '''file exists'''
                if isBuildServer():
                    '''no need to create another file'''
                    create_dist_pkg = False
                else:
                    '''check timestamp for the last modified tag (including subdirectories)'''
                    last_modified = time.localtime(max(os.path.getmtime(root) for root, _, _ in os.walk(pkg_path)))
                    create_dist_pkg = last_modified > time.localtime(os.path.getmtime(pkg_dist))
            else:
                '''file does not exist yet'''
                create_dist_pkg = True

            if create_dist_pkg:
                pkg_dist = DistPkgUtil.createFromPkgDir(pkg_path, self.py_pkg_file_format, pkg_dist_dir,
                                                        pkg_version=pkg_version, compile_bytecode=True)
            copy_dist_pkgs.append(pkg_dist)
            if pkg_name in std_pkgs:
                '''only add versions of standard packages'''
                pkg_versions[pkg_name] = pkg_version

        copy_unique_modules = list(set(copy_modules))
        copy_unique_dist_pkgs = list(set(copy_dist_pkgs))

        if bool(pkg_versions):
            '''write a json file, which lists all packages and versions used '''
            pkgs_json = ComponentPaths.getPackagesFile(self.path)
            with open(pkgs_json, 'w') as fh:
                json.dump(pkg_versions, fh, indent=4, sort_keys=True)
        return copy_unique_modules, copy_unique_dist_pkgs

    @staticmethod
    def __addSyMFile(comp: SymComponent, c_folder: str, comp_files: list, file_path: str):
        """
        Add (copy and register/link) additional files to the component

        Parameters
        ----------
        comp : SymComponent
            Component Object
        c_folder : string
            Path to component files
        comp_files : list
            Component files list
        file_path : string
            Path to the file

        """
        # Todo: make this simpler -> reduce method parameters etc...
        if not os.path.isfile(file_path):
            raise FileNotFoundError('File does not exist: {}'.format(file_path))

        new_file = cs_fileutils.copyFile(file_path, c_folder)
        comp.addLinkedFile(new_file)
        print('... successfully copied {}'.format(file_path))

        ''' Replace file names '''
        for cf in comp_files:
            if cf.getData() == file_path:
                cf.setData(new_file)
                comp.addLinkedFile(new_file)
                print('... ... re-linking SyMFile {} to {}'.format(cf.getName(),
                                                                   os.path.relpath(new_file, os.path.dirname(c_folder))
                                                                   )
                      )

    @staticmethod
    def __addSyMDirectory(comp: SymComponent, p_folder: str, c_folder: str, comp_files: list, comp_dirs: list,
                          sdir: str):
        """
        Add (copy and register/link) additional directories to the component

        Parameters
        ----------
        comp : SymComponent
            Component Object
        p_folder : str
            Path to project file
        c_folder : str
            Path to component files
        comp_files : list
            list of files, which will be copied into the component
        comp_dirs : list
            list of directories, which will be copied into the component
        sdir : str
            Path to the source directory

        """
        # Todo: make this simpler -> reduce method parameters etc...
        dir_name = os.path.basename(os.path.normpath(sdir))
        if dir_name.find('_') == 0:
            ddir = os.path.join(p_folder, dir_name)
        else:
            ddir = os.path.join(c_folder, dir_name)

        ''' copy all files and folders in sdir to ddir'''
        print('... copy directory {}'.format(sdir))
        cs_fileutils.copyTree(sdir, ddir)

        ''' register all files in ddir'''
        hidden_paths = glob.glob(ddir + '/**/.*', recursive=True)
        paths = glob.glob(ddir + '/**/*', recursive=True)
        paths.extend(hidden_paths)
        files = set([path for path in paths if os.path.isfile(path)])

        for file in files:
            comp.addLinkedFile(file)
            print('... ... successfully copied {}'.format(os.path.relpath(file, ddir)))
            for cf in comp_files:
                if cf.getData().startswith(sdir):
                    cf.setData(cf.getData().replace(sdir, ddir))
                    print('... ... re-linking SyMFile {}'.format(cf.getName()))

        ''' Replace file names '''
        for cd in comp_dirs:
            if cd.getData().find(sdir) == 0:
                cd.setData(ddir)
                print('... ... ... re-linking SyMDirectory {} to {}'.format(cd.getName(), cd.getData()))

    def _strMopFile(self):
        mopfile = os.path.join(self.path, self.name + '.mop')
        sym_home = Paths.getSyMSpaceRoot()
        if not sym_home.endswith('\\'):
            sym_home = sym_home + '\\'
        comp_home = Paths.getComponentSpaceRoot()
        if not comp_home.endswith('\\'):
            comp_home = comp_home + '\\'

        if mopfile.startswith(sym_home):
            path_lst = mopfile.replace(sym_home, '').split('\\')
            path = ', '.join(["'" + p + "'" for p in path_lst])
            return "os.path.join(os.getenv('MagOpt_Home'), " + path + ")"

        elif mopfile.startswith(comp_home):
            path_lst = mopfile.replace(comp_home, '').split('\\')
            path = ', '.join(["'" + p + "'" for p in path_lst])
            return "os.path.join(os.getenv('ComponentSpace'), " + path + ")"
        else:
            path = mopfile.replace('\\', '\\\\')
            return "'{}'".format(path)

    def buildTest(self, add_component: bool = True, add_testparam: bool = True) -> SymProject:
        """
        Creates a testproject and saves it to the filesystem

        Parameters
        ----------
        add_component : bool
            Add the component to the testproject, True (Default)
        add_testparam : bool
            Copy the Testparameter Container to the testproject, True (Default)

        Returns
        -------
        test_sym_mdl : SymProject
            testproject
        """
        self.printH2('Build Test')

        d = self.preBuildTest()

        test_mop_file = d['MopFile']
        comp_webpath = d['WebPath']
        testparam = d['TestParam']

        if os.path.isfile(test_mop_file + ".mop"):
            os.remove(test_mop_file + ".mop")
        if os.path.isdir(test_mop_file + ".mop.Files"):
            shutil.rmtree(test_mop_file + ".mop.Files")

        prj_name = 'Test' + self.name.replace('.', '_')
        test_sym_mdl = SymProject.new(prj_name, '', test_mop_file)

        if add_component:
            '''add the Component to the Test-Project'''
            test_sym_mdl.addComponent(comp_webpath)
        if add_testparam:
            '''add the Testparameter container to the Test-Project'''
            test_sym_mdl.addParam(testparam)

            '''collect unittest reference parameters (_test_*) and move them to the correct position'''
            test_params = test_sym_mdl['Testparameter'].getParameters('SymField')

            for test_param in test_params:
                if test_param.getName().startswith('_test_'):
                    new_id = test_param.getIdentifier().replace('.Testparameter', '')
                    cont_id = '.'.join(new_id.split('.')[:-1])

                    if test_sym_mdl.getUniqueObject(cont_id) is not None:
                        '''find the container element with the exact/unique identifier'''
                        test_sym_mdl[cont_id].addParam(test_param)
                    else:
                        raise Exception('Parameter ID does not exist: {}'.format(cont_id))

            '''collect files in testparameters and adjust the link'''
            test_files = test_sym_mdl['Testparameter'].getParameters('SymFile')
            for test_file in test_files:
                test_mop_path = test_sym_mdl.getAllFilesDirectory()
                test_file.setAbsoluteFilePath(os.path.join(os.path.join(test_mop_path, '_CompFiles'),
                                                           test_file.getFileName()))

        test_sym_mdl.save()
        test_sym_mdl.close()

        print('Component test project saved to: {}'.format(test_mop_file))

        return test_sym_mdl

    def preBuildTest(self) -> dict:
        """
        Get parameters for test file

        Returns
        -------
        d : dict
        """
        mop_file = os.path.join(self.path, self.name)
        if not os.path.isfile(mop_file) and not os.path.isfile(mop_file + ".mop"):
            raise FileNotFoundError(mop_file + " can not be found.")
        sym_mdl = SymProject.load(mop_file)
        if not isinstance(sym_mdl, SymProject):
            raise Exception(mop_file + " can not be loaded.")

        webpath = sym_mdl['Info']['WebPath'].getData()
        name = sym_mdl['Name'].getData()
        comp_webpath = '.'.join([webpath, name])

        testparam = sym_mdl['Testparameter']

        sym_mdl.close()

        if testparam is None:
            raise Exception('Container Testparameter does not exist!')

        test_mop_file = os.path.join(self.base_directory, 'tests', 'test_{}'.format(self.name))

        d = {'MopFile': test_mop_file,
             'WebPath': comp_webpath,
             'TestParam': testparam}
        return d

    def writeJSON(self, fname_comp: str, help_dir: str, **kwargs):
        """
        Write JSON File

        Parameters
        ----------
        fname_comp : str
            File name of the Component.
        help_dir : str

        """
        prj = self.sym_mdl
        inc_lst = ['Info', 'Files', 'Docu', 'Filter']
        prj_dict = prj.toDict()
        json_dict = {}

        for item in inc_lst:
            if item in prj_dict:
                json_dict[item] = prj_dict[item]

        # Define type if not specified in Filter
        if 'type' not in json_dict['Filter']:
            json_dict['Filter']['type'] = self.name

        # Infos for database
        types_enum = {'OTHER': '0', 'PROPERTY': '1', 'PARAMETER': '2', 'TAG': '3'}

        # Redefine provides
        ruleCond = []
        if 'provides' in json_dict['Filter']:
            provides = json_dict['Filter']['provides']

            for key in provides:
                if isinstance(provides[key], dict):
                    cond = provides[key]
                    if 'type' in cond:
                        cond['type'] = types_enum[cond['type']]
                    cond.pop('__error__', None)
                    rule = {'logicalOperator': 'OR',
                            'ruleConditions': [cond],
                            'name': cond['value']}
                    if 'guid' in cond:
                        rule['guid'] = cond['guid']
                        cond.pop('guid', None)
                    if 'displayName' in cond:
                        rule['name'] = cond['displayName']
                        cond.pop('displayName', None)
                    ruleCond.append(rule)

            provisions = {'logicalOperator': 'OR', 'rules': ruleCond}
            json_dict['Filter']['provisions'] = provisions

            ''' Remove provides from json_dict '''
            json_dict['Filter'].pop('provides', None)

        ''' Redefine requires '''
        rule = []
        if 'requires' in json_dict['Filter']:
            requires = json_dict['Filter']['requires']

            if 'AND' in requires:
                operator = 'AND'
                subop = requires['AND']
            elif 'OR' in requires:
                operator = 'OR'
                subop = requires['OR']
            else:
                operator = None
                subop = []

            for op in subop:
                innerop = subop[op]
                if not isinstance(innerop, dict):
                    continue

                if op.startswith('AND'):
                    log_operator = 'AND'
                elif op.startswith('OR'):
                    log_operator = 'OR'
                else:
                    raise Exception('Unexpected case! logical operator={}'.format(op))

                ruleCond = []
                ruleInfo = {'logicalOperator': log_operator}

                displayName = op.replace(log_operator, '')
                if displayName.startswith('_'):
                    displayName = displayName[1:]
                ruleInfo['name'] = '"{}"'.format(displayName)

                for key in innerop:
                    if isinstance(innerop[key], dict):
                        cond = innerop[key]
                        if 'type' in cond:
                            cond['type'] = types_enum[cond['type']]
                        cond.pop('__error__', None)
                        ruleCond.append(cond)
                    else:
                        if key == 'min':
                            ruleInfo['min'] = '{}'.format(int(float(innerop[key])))
                        elif key == 'max':
                            ruleInfo['max'] = '{}'.format(int(float(innerop[key])))
                        elif key == 'displayName':
                            ruleInfo['name'] = innerop[key]
                        elif key == '__error__':
                            pass
                        elif key == 'guid':
                            ruleInfo['guid'] = innerop[key]
                        elif key == 'assemblyRules':
                            ruleInfo['assemblyRules'] = json.loads(innerop[key])
                        elif key == 'optional':
                            ruleInfo['optional'] = innerop[key]
                        else:
                            raise Exception('Key Error "{}" in {}'.format(key, innerop))
                ruleInfo['ruleConditions'] = ruleCond
                rule.append(ruleInfo)

            if operator is not None:
                requirements = {'logicalOperator': operator, 'rules': rule}
                json_dict['Filter']['requirements'] = requirements

            ''' Remove requires from json_dict '''
            json_dict['Filter'].pop('requires', None)

        '''Add help directory'''
        json_dict['Filter']['HelpDir'] = help_dir

        '''Add UserGroup and User'''
        if 'User' in json_dict['Filter'] and not prj.getObject('Filter.User').getData():
            json_dict['Filter'].pop('User', None)

        '''Add kwargs'''
        for key in kwargs:
            json_dict[key] = kwargs[key]

        fname = fname_comp.replace('.mopc', '.json')

        # Todo: should all absolute paths matching the environment variable ComponentSpace
        #  e.g. "C:\ComponentSpace\trunk" or "C:\\opt\\Jenkins\\jobs\\ComponentSpace\\workspace\\ComponentSpace"
        #  be replaced with a unambiguous placeholder  e.g. %ComponentSpace% or ${ComponentSpace}
        with open(fname, 'w+') as fh:
            '''pretty print json'''
            json.dump(json_dict, fh, indent=4, sort_keys=True)

    @staticmethod
    def getStructure(container):
        cont_type = str(type(container)).split('.')[-1].replace("'>", '')
        structure = {'name': container.getName(),
                     'description': container.getDescription(),
                     'access': container.getAccess(),
                     'type': cont_type}
        param = []
        for p in container.getParam():
            if isinstance(p, SymContainer):
                # Add public interface
                parent = p.getParent().getIdentifier() + '.'
                for name in p.listPublicInterfaceName():
                    if p.getPublicInterfaceParameter(name) is not None:
                        link = p.getPublicInterfaceParameter(name).getIdentifier().replace(parent, '')
                        param.append({'name': name,
                                      'type': 'Interface',
                                      'link': link})

                param.append(Component.getStructure(p))
            else:
                name = p.getName()
                desc = p.getDescription()
                par_type = str(type(p)).split('.')[-1].replace("'>", '')
                if hasattr(p, 'getUnit'):
                    # todo: Change to standardized SI units
                    unit = p.getUnit()
                    siunit = ''
                else:
                    unit = ''
                    siunit = ''
                if hasattr(p, 'getData') and hasattr(p, 'getLock') and p.getLock():
                    value = p.getData()
                    if not isinstance(value, (str, float)):
                        value = None
                else:
                    value = None
                par_info = dict(name=name, description=desc, type=par_type, siunit=siunit)
                if value is not None:
                    par_info.update(dict(value=value, unit=unit))

                param.append(par_info)

        structure['parameters'] = param
        return structure

    @staticmethod
    def getUnresolvedInterface(container):
        root = container.getRoot().getName()
        imports = []
        for par in container.getUnresolvedImports():
            imports.append({'parameter': par[0].getIdentifier().replace(root + '.', ''),
                            'link': par[1].getParamName(),
                            'siunit': ''})

        exports = []
        for par in container.getUnresolvedExports():
            exports.append({'parameter': par[0].getIdentifier().replace(root + '.', ''),
                            'link': par[1].getParamName(),
                            'siunit': ''})

        return {'imports': imports, 'exports': exports}

    def getDependencies(self, container: SymContainer, webpath: str) -> dict:
        """
        Collect dependency on other components

        Parameters
        ----------
        container : SymContainer
        webpath: str
            Webpath of the component

        Returns
        -------
        dict
        """

        webpaths = []

        for sub_param in container.getParameters('SymContainer'):
            if sub_param.isComponent() and not isinstance(sub_param, SymProject):
                '''check if component, but skip basemodel'''
                # sub_webpath = sub_param.getComment()
                # print(sub_param.getIdentifier())
                sub_webpath = sub_param.getComponentName()
                webpaths.append(sub_webpath)

        dependencies = []

        for sub_webpath in webpaths:
            if sub_webpath is None or sub_webpath == "":
                raise Exception('Empty Comment field => Component Webpath is missing')
            elif sub_webpath != webpath:
                '''add only other webpaths to dependencies'''
                dependencies.append(sub_webpath)

        '''filter duplicate webpaths'''
        dependencies = list(set(dependencies))
        return {'base_directory': self.base_directory.replace(Paths.getComponentSpaceRoot(), '%ComponentSpace%'),
                'webpaths': dependencies}
